// This is ds_expt (c) Matt Clark September 2002
// This release is covered by GPL - it is free software 
//
// This program is a synthetic hologram or computer designed
// hologram design program.  It implements an algorithm described
// in Applied Optics page 5331, volume 38, number 25 1999
// The full bibtex reference is below.  This algorithm is 
// in turn based on the state-variables direct search method
// described in Optics Communications, page 150, volume 124,
// number 1-2, 1996.  If you publish work based on this algorithm
// please cite these two papers rather than the program.
// The program can implement either algorithm.
// 
// Thanks matt
//
// @article{ds_expt,
//         title="Two-Dimensional, Three-Dimensional, and Gray-Scale Images Reconstructed from Computer-Generated Holograms Designed by use of a Direct-Search Method ",
//         author = "Matthew Clark",
//         journal="Applied Optics",
//         volume= "38",
//         number="25",
//         pages="5331-5337",
//         year=1999
//         }
//
// @article{ds_sv,
//         title="A direct-search method for the computer design of holograms",
//         author="Matthew Clark and Robin Smith",
//         journal =       "Optics Communications",
//         volume= "124",
//         number="1-2",
//         pages="150-164",
//         year=1996
//         }
//
// The image points must be separated in space according to the sampling conditions 
// specified in  App Opt 38(25) p5331 1999.  Recapped here - (1) workout the
// point spread function size for your hologram (wavelength / NA). (2) Divide this
// distance by 2. (3) this is the sampling distance required for continuous 
// reconstruction of the image, (4) To get discrete dots use at least 2x (1).
// Inbetween these limits you may get partially joined lines / dots
//
// Iff you must have discrete dots between these two limits you need to put a 
// dark dot between the bright ones you want.
//


/*  This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// c headers
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <stdlib.h>
#include <signal.h>
// c++ headers
#include <iostream>
#include <fstream>
#include <complex>
// own headers
#include "ds_expt.h"

using namespace std;

#define MAX_NAME 256
#define NULL_PHASE -1
#define SMALL_NUMBER 1e-6

/* global declarations */
BOOL		continue_ok=TRUE;
BOOL		dump_data=FALSE;

double           wavelength=633e-9;
int		max_sweeps=100;

// here are the global cost balancing factors
double          a=2,b=1;

inline double    phase(double range){
	return(range*K);
	}

inline double rdist(double xi, double yi, double x, double y,double dz){
	double            dx, dy;
	dx = xi - x;
	dy = yi - y;
	return (sqrt(dz * dz + dx * dx + dy * dy));
	}

inline int randnum(long);

void	p(char *temp){
	printf("%s\n",temp);
	}

inline double cfr(double,double,double,double);

// PURE-base-base-class-for-all-optics-objects

typedef class Optics_Object{
	public:
		Optics_Object(){next=NULL;};
		Optics_Object(ifstream *is){loadset(is);next=NULL;};
		virtual	void		loadset(char *){};
		virtual	void		loadset(ifstream *){};
		virtual	void	restore(char *){};
		virtual	void	restore(ifstream *){};
		virtual	void	save(char *){};
		virtual	void	save(ifstream *){};
		void	*	next;
	} Optics_Object;

// -----------------------------Hologram object declarations-- -- -------
typedef class Hologram:public Optics_Object {
	private:
		BOOL		aperture_on;
		BOOL		circ_ap;
		float           prob_accept,hx,hy;//center coords for hologram
		long            no_examined,no_accepted;
		//used to monitor the probability of change between report periods
		long            rep_examined,rep_accepted;
		long		list_pointer;
		// int points for the exhaustive list implementation
		int		*list_x,*list_y;
		int		plist_x,plist_y;
		
		int		sample_size;
		char		aperture_file[MAX_NAME];
		char            holname[MAX_NAME];
		BOOL            test_nearest(int,int);
		void		setup_exhaustive_list();
		HOLODATA        data;
		float           phase_constant,stability;
		int 		ihx,ihy;//integer center coordinates
		long            reg,no_points;
		double		magconst;
	public:
		int		sweep;
		char  		newphase,oldphase;
		CHANGE_T        generate_change(void);
		void	           keep_change(void);
		void	           loaddata(void);
		void	           savedata(void);
		void	           do_report(ofstream *);
		BOOL            stable_solution();
		Hologram(ifstream *);
	} Hologram;



inline void	Hologram::do_report(ofstream *os){
float	temp;

	temp=(((float)rep_accepted)/((float)rep_examined));
	//if(temp>1)temp=1;
	(*os)<<" "<<temp;
	rep_accepted=1;rep_examined=1;
	}

inline BOOL Hologram::stable_solution(void){
BOOL ret=FALSE;

	if(prob_accept<stability)ret=TRUE;
	return ret;
	}

void	Hologram::loaddata(void){
ifstream is;
int i,ph;

	for(i=0;i<no_points;i++){
		data.hdata[i]=NULL_PHASE;
		}
	if(aperture_on==TRUE){
       		is.open(aperture_file);
        	if(is.bad())p("aperture file load error");
        	for(i=0;i<no_points;i++){
               		is>>ph;
               		data.aperture[i]=ph;
                	}
        	}
	}

void	Hologram::savedata(void){
FILE *fo;

	fo=fopen(holname,"w");
	fwrite((void	*)data.hdata,sizeof(HDATA),data.xpixels*data.ypixels,fo);
	fclose(fo);
	}


inline void	Hologram::keep_change(void){
	data.hdata[reg]=newphase;
	no_accepted++;
	rep_accepted++;
	}



Hologram::Hologram(ifstream *is){
char temp[MAX_STRING];
int tint,seed,counter=0;

	//set up the hologram variables to their default values
	newphase=0;
	oldphase=0;
	stability=1;
	wavelength=633e-9;
	prob_accept=1;
	rep_examined=1;// to prevent / by zero
	rep_accepted=1;// to prevent / by zero
	no_accepted=1;//to prevent / by zero;
	no_examined=1;//to prevent / by zero;
	data.phaselevels=2;
	list_pointer=0;
	sweep=0;
	plist_x=0;
	plist_y=0;
	aperture_on=FALSE;
	circ_ap=FALSE;
	sample_size=10000;//how many trials before stats are renewed;
	//check end of this section for bug fix 1997 cf sample_size
	//Now readin the hologram options
	do{
		(*is)>>temp;
		counter++;
		if(strcmp(temp,"aperture_file")==0){
			aperture_on=TRUE;
			(*is)>>aperture_file;
			}
		if(strcmp(temp,"circular_aperture")==0){
			circ_ap=TRUE;
			cout<<"Aperture set to circular\n";
			}
		if(strcmp(temp,"phase") == 0) {
			(*is)>>data.phaselevels;
			}
		if (strcmp(temp, "hologram") == 0) {
			(*is)>>holname;
			}
		if (strcmp(temp, "pixels") == 0) {
			(*is)>>data.xpixels>>data.ypixels;
			no_points=data.xpixels*data.ypixels;
			cout<<no_points<<" hologram samples in total\n";}
		if (strcmp(temp, "psize") == 0) {
			(*is)>>data.pixelsize.x>>data.pixelsize.y;
			}
		if (strcmp(temp, "wavelength") == 0) {
			(*is)>>wavelength;
			}
		if (strcmp(temp, "beamtype") == 0) {
			(*is)>>data.beam_type;
			}
		if (strcmp(temp, "beamdata") == 0) {
			(*is)>>data.beamdata;
			}
		if(strcmp(temp, "stable_at") == 0) {
			(*is)>>stability;
			}
		}
	while (strcmp(temp, "endhologram") != 0&&counter<MAX_OPTIONS);
	//check for an hologram options read error
	if(counter==MAX_OPTIONS){
		cout<<"hologram option read error \n";
		exit(1);
		}

	cout<<"beam type "<<data.beam_type<<"\n";
	cout<<"current beam types are "<<plane<<" plane, "<<gaussian<<" gaussian\n";
	phase_constant=PIX2/(float)data.phaselevels;
	cout<<"We are stable at "<<stability<<endl;
	//total number of pixels
	//allocate hologram memory
	no_points=data.xpixels*data.ypixels;
	data.hdata=new HDATA[no_points];
	if(aperture_on==TRUE){
		data.aperture=new int[no_points];
		}
	if(data.hdata==NULL){
		cout<<"Fatal error:Unable to allocate hdata\n";
        	cout<<"Requires "<<no_points<<" units sizeof "<<sizeof(HDATA)<<"\n";
		p("Hologram mem alloc err");
		exit(3);	
		}
	magconst=(2.0/wavelength)*(data.pixelsize.x*data.pixelsize.y);
	hx=data.xpixels*data.pixelsize.x/2;
	hy=data.ypixels*data.pixelsize.y/2;
	ihx=data.xpixels/2;
	ihy=data.ypixels/2;
	cout <<data.pixelsize.x<<"   "<<data.pixelsize.y<<" \n";
	//check allocate and set up the exhaustive list option
	cout<<"setting up exhaustive list\n";
	list_x=new int[data.xpixels];
	list_y=new int[data.ypixels];
	if(list_x==NULL||list_y==NULL){
       		cout<<"Fatal error:Unable to allocate exhaustive_list\n";
		cout<<"Requires "<<data.xpixels+data.ypixels<<" units sizeof "<<sizeof(int)<<"\n";
        	p("Hologram mem alloc err:exhaustive list");
        	exit(3);
        	}
	setup_exhaustive_list();
	if(sample_size<(no_points/2)){sample_size=no_points/2;}
	}

inline int randnum(long top){
	float temp;
	temp=rand();
	temp*=top;
	temp/=RAND_MAX;
	return (int)temp;
	}



void	Hologram::setup_exhaustive_list(void){
	// to produce a random exhaustive list of the
	// pixels,each visited exactly once in the list
	// Important change in DS_RE
	// random list implemented differently using separate x and y lists
	// to conserve memory may 1996
	
	int i=0,p;
	int temp;
	//set sequential x and y lists
	for(i=0;i<data.xpixels;i++){
		list_x[i]=i;
		}
	for(i=0;i<data.ypixels;i++){
		list_y[i]=i;
		}
	// scramble x and y lists
	for(i=0;i<data.xpixels;i++){
		p=randnum(data.xpixels);
		temp=list_x[i];
		list_x[i]=list_x[p];
		list_x[p]=temp;
		}
	for(i=0;i<data.ypixels;i++){
	        p=randnum(data.ypixels);
	        temp=list_y[i];
	        list_y[i]=list_y[p];
	        list_y[p]=temp;
	        }
	
	sweep++;
	cout<<"Sweep level now "<<sweep<<endl;
	}

CHANGE_T Hologram::generate_change(void){
CHANGE_T        change;
BOOL		pixel_ok;
int		px,py;	

	change.mag=0;
	
	// check the stats and set the sampling period 
	if(no_examined>sample_size)
		{
		prob_accept=(float)no_accepted/(float)no_examined;
	 	cout<<"prob="<<prob_accept<<"\n";
                no_examined=1;no_accepted=1;
		}

	//Select the next pixel using the RANDOM, EXHAUSTIVE technique with independant
	//recycled separate scrambled lists for x and y coordinates (to save memory)
	do{//test nearest loop
		do{//select the pixel and check it is in the aperture
			pixel_ok=TRUE;
			px=list_x[plist_x++];
			py=list_y[plist_y++];
			if(plist_x>=data.xpixels){plist_x=0;plist_y--;};
			if(plist_y>=data.ypixels)plist_y=0;	
			if(++list_pointer>no_points){
				list_pointer=0;
				setup_exhaustive_list();
				}
			reg=px + data.xpixels * py;
			// 	test the pixel it is ok if
			// 	the aperture_on and circ_ap are FALSE
			// OR	aperture_on is TRUE and data.aperture[reg]!=APERTURE_OFF
			// OR	circ_ap is TRUE and R(reg)<ihx	
			// not all of these test can be done at the same time because of different data availability
			// so we will use a pixel_ok flag (not too elegant BUT tranparent, should functionalise)
			// A pixel is NOT ok if it is not on an edge

		if(aperture_on==TRUE){
			if(data.aperture[reg]==APERTURE_OFF)pixel_ok==FALSE;
			}
			else{
			if(circ_ap==TRUE){
                   		int x,y;
                        	x=px-ihx;
                        	y=py-ihy;
                        	if(sqrt((double)x*(double)x+(double)y*(double)y)>(float)ihx)pixel_ok=FALSE;
                                	}
				}
			}
			while(pixel_ok==FALSE);
		//if it is acceptable for the above regions it counts as 'tested'
		no_examined++;
		}
		while(test_nearest(px,py)==FALSE);
	//Now the pixel selected is in the aperture and not in a fringe boundary
	//Now select the phase, making sure it is NOT the same as the existing level
	if(data.hdata[reg]==NULL_PHASE){
	//now choose from any of the levels
		newphase=randnum((long)data.phaselevels);
		}
		else{
			if(data.phaselevels==2){
	//there is only one choice, try zero first
				newphase=0;
				}
			else{
	//there is n-1 choices, select one of those
				newphase=randnum((long)data.phaselevels-1);
				}
	//now test to see if we have the current level, if so choose the last level (because its not the current one)
			if(data.hdata[reg]==newphase) newphase=data.phaselevels-1;
			}
		rep_examined++;

	//Now convert that change into useful data
		change.pos.x= ((double)(px) * data.pixelsize.x)-hx;
		change.pos.y= ((double)(py) * data.pixelsize.y)-hy;
		oldphase = data.hdata[reg];
		change.oldphase=(double)oldphase*phase_constant;
		change.newphase=(double)newphase*phase_constant;
		change.newphaselevel=newphase;
		change.oldphaselevel=oldphase;
		if(data.phaselevels==2){
			change.binary=TRUE;
			}
			else{
			change.binary=FALSE;
			}
		switch(data.beam_type){
       			case plane:
                		change.mag=magconst;
                		break;
        		case gaussian:
				float r,x,y;
                		x=change.pos.x;
				y=change.pos.y;
				r=sqrt(x*x+y*y);
				change.mag=magconst*exp(-((r/hx)*(r/hx)));
				//magnitude of gaussian, 1/e width = 1./2 hologram size
                		break;
        		default: 
				printf("Unknown beamtype specified\n");
				exit(1);
                		break;
        		}
		return change;
		}
	
BOOL Hologram::test_nearest(int px,int py){
/* this test to see if the nearest neighbours are different */
int	phase,temp = 0;

	reg=px + data.xpixels * py;
	phase=data.hdata[reg];
	if(phase==NULL_PHASE)return TRUE;//null phase pixels

	if (px > 0 && (phase == data.hdata[reg - 1]))
		temp++;
	if (px < (data.xpixels - 1) && (phase == data.hdata[reg + 1]))
		temp++;
	if (py > 0 && (phase == data.hdata[reg - data.xpixels]))
		temp++;
	if (py < (data.ypixels - 1) && (phase == data.hdata[reg + data.xpixels]))
		temp++;
	if(temp==4)return FALSE;
	return TRUE;
	}


//-IP--------------------------image part class type-- -- --------------

typedef class Image_Object:public Optics_Object {
	public:
		complex <CGH_TYPE>		*newdata,*olddata;
		//now declare data
		char            posname[MAX_NAME], ampname[MAX_NAME];
		COST_T          costs;
		// now start methods    
		virtual void	           loaddata(void){};
		virtual void	           calc_change(CHANGE_T){};
		virtual void	           calc_change_bin(CHANGE_T){};
		virtual void	           intro_change(CHANGE_T){};
		virtual void	           savedata(void){};
		virtual void	           keep_change(void);
		virtual COST_T          do_change(CHANGE_T);
		virtual void	           do_report(ofstream *os);
		virtual double          aver_int(void){return -1;};
		Image_Object(){	costs.newc.signal=0;costs.newc.noise=0;
				costs.oldc.signal=0;costs.oldc.noise=1e30;
				next=NULL;};
	} Image_Object;

void	Image_Object::keep_change(){
complex <CGH_TYPE> *temp;

	//swap the image data about
	temp       =       newdata;
	newdata    =       olddata;
	olddata    =       temp;
	//now update the costs
	costs.oldc.signal =       costs.newc.signal;
	costs.oldc.noise =        costs.newc.noise;
	}

inline void	Image_Object::do_report(ofstream *os){
	(*os)<<" "<<costs.oldc.signal<<" "<<costs.oldc.noise<<" ";
	}

//--COSTFUNCTION-----------------------------------------------------

COST_T Image_Object::do_change(CHANGE_T change){
        if(change.oldphaselevel==NULL_PHASE){
                intro_change(change);
                }
        else{   
		if(change.binary==TRUE){
			calc_change_bin(change);
			}
		else{
			calc_change(change);
			}
                }
	return costs;
	}

inline double cfr(double i1,double i2,double t,double n){
	return SMALL_NUMBER+i2-2*t*i1+t*t*n;
	}

//----Object2d----------------------plane-2-d-object-with-uniform-intensity-----
// this forms the basis for most image objects, the functions could be 
// rationalised or modulorised BUT these are extremely perfomance sensitive
// and it is easier for them to be optimized if they are as plain as possible
// hence writing them out and a lot of repeating myself

typedef class Object2d : public Image_Object {
	private:
		double		distance;
	public:
		POINT		*position;
		int		no_points;
		void	           loaddata(void);
		void	           calc_change(CHANGE_T);
		void	           calc_change_bin(CHANGE_T);
		void	           intro_change(CHANGE_T); 
		void	           savedata(void);
		Object2d(ifstream *is);
		Object2d(){};//default constructor
		// The cost parameters are kept raw here for use by the Image object
		// it consists of a COST_T structure
	} Object2d;

void	Object2d::intro_change(CHANGE_T ch){
int             i;
double         mag;
double         r,p;
double         I,sigmaI=0,sigmaI2=0,Ibar=0;
complex <CGH_TYPE> 	amp;

	for (i = 0; i < no_points; i++) {
		mag=ch.mag/(r=rdist(position[i].x, position[i].y, ch.pos.x, ch.pos.y, distance));
		//this means a positive phase is equivalent to extra path length
		p = phase(r) + ch.newphase;
		amp=newdata[i]=olddata[i]+polar(mag,p);
		I=real(amp*conj(amp));
		sigmaI+=I;             //sum the intensity
		sigmaI2+=(I*I);        //sum of intensity ^2;
		}
	Ibar=costs.newc.signal=sigmaI/no_points;
	costs.newc.noise=sqrt(((sigmaI2-2*Ibar*sigmaI)/no_points)+Ibar*Ibar);
	}

void	Object2d::calc_change_bin(CHANGE_T ch){
int             i;
double         	A,mag,temp;
double         	r,p;
double         	I,sigmaI=0,sigmaI2=0,Ibar=0;
complex <CGH_TYPE> 	amp;

	//Binary phase version a factor of 2 in the mag
	A=ch.mag*2;
	for (i = 0; i < no_points; i++) {
        	mag =A  / (r = rdist(position[i].x, position[i].y, ch.pos.x, ch.pos.y, distance));
	//Then only require 1/2 as many sines and cosines
		p = phase(r) + ch.newphase;
		amp=newdata[i]=olddata[i]+polar(mag,p);
		I=real(amp*conj(amp));
		sigmaI+=I;             //sum the intensity
		sigmaI2+=(I*I);        //sum of intensity ^2;
        	}
	Ibar=costs.newc.signal=sigmaI/no_points;
	costs.newc.noise=sqrt(((sigmaI2-2*Ibar*sigmaI)/no_points)+Ibar*Ibar);
	}

void	Object2d::calc_change(CHANGE_T ch){
int             i;
double         mag,temp;
double         r,p1,p2;
double         I,sigmaI=0,sigmaI2=0,Ibar=0;
complex <CGH_TYPE>  	amp;

	//substitue in to allow vectorization
	for (i = 0; i < no_points; i++) {
		mag =ch.mag  / (r = rdist(position[i].x, position[i].y, ch.pos.x, ch.pos.y, distance));
		temp = phase(r);
		p1 = temp+ch.newphase;
		p2 = temp+ch.oldphase;
		amp=newdata[i]=olddata[i]+polar(mag,p1)-polar(mag,p2);
		// The previous line MAY be inefficient...... matt 11/Aug/98
		I=real(amp*conj(amp));
		sigmaI+=I;             //sum the intensity
		sigmaI2+=(I*I);        //sum of intensity ^2;
		}
	Ibar=costs.newc.signal=sigmaI/no_points;
	costs.newc.noise=sqrt(((sigmaI2-2*Ibar*sigmaI)/no_points)+Ibar*Ibar);
	}
 
 
Object2d::Object2d(ifstream * is){
char temp[MAX_STRING],phase_table_name[MAX_STRING];
int counter=0;

	(*is)>>posname>>ampname;
	cout<<"pos name="<<posname<<" amp name="<<ampname<<"\n";
	do {
		(*is)>>temp;
		counter++;
		if (strcmp(temp, "imagesize") == 0) {
			(*is)>>no_points;
			}
		if (strcmp(temp, "distance") == 0) {
			(*is)>>distance;
			}
		}while (strcmp(temp, "end") != 0&&counter<MAX_OPTIONS);
	if(counter==MAX_OPTIONS){
        	cout<<"2dOjbect option read error \n";
        	exit(1);
        	}
	
	position=new POINT[no_points];
	newdata=new complex <CGH_TYPE>[no_points];
	olddata=new complex <CGH_TYPE>[no_points];
	if(position==NULL||newdata==NULL||olddata==NULL)p("line mem alloc err");
	for(int i=0;i<no_points;i++){
        	newdata[i]=complex <CGH_TYPE>(0,0);
        	olddata[i]=complex <CGH_TYPE>(0,0);
		}
	}

void	Object2d::loaddata(void){
int i;

	ifstream is;
	is.open(posname);
	if(is.bad()){	cout<<"2dObject .pos data no good exiting \n";
			exit(3);
			}
	for(i=0;i<no_points;i++){
		is>>(position[i].x)>>(position[i].y);
		}
	is.close();
	}


void	Object2d::savedata(void){
FILE *fo;

	fo=fopen(ampname,"w");
	fwrite((void	*)newdata,sizeof(complex<CGH_TYPE>),no_points,fo);
	fclose(fo);
	}


// now Object3d based on object2d, main difference is that the .pos
// file contains x y z coordinates and that each point has a different z
// now of course this could be general with the same z for 2d stuff but
// it is again an optimisation issue, however I never did test the speed
// difference and i think it would be very minimal on most platforms, its
// main imapact is to increase the memory bandwidth requirement by about
// 1/6th, probably no problem as this is SO cpu limited

typedef class Object3d : public Object2d {
	private:
		float		*depth;
	public:
		void	           loaddata(void);
		void	           calc_change(CHANGE_T);
		void	           calc_change_bin(CHANGE_T);
		void	           intro_change(CHANGE_T); 
		Object3d(ifstream *is);
		Object3d(){};//default constructor
		// The cost parameters are kept raw here for use by the Image object
		// it consists of a COST_T structure
	} Object3d;



void	Object3d::intro_change(CHANGE_T ch){
int             i;
double         mag;
double         r,p;
double         I,sigmaI=0,sigmaI2=0,Ibar=0;
complex <CGH_TYPE>		amp;

	for (i = 0; i < no_points; i++) {
		mag =ch.mag  / (r = rdist(position[i].x, position[i].y, ch.pos.x, ch.pos.y, depth[i]));
		//this means a positive phase is equivalent to extra path length
		p = phase(r) + ch.newphase;
		amp=newdata[i]=olddata[i]+polar(mag,p);
		I=real(amp*conj(amp));
		sigmaI+=I;             //sum the intensity
		sigmaI2+=(I*I);        //sum of intensity ^2;
		}
	Ibar=costs.newc.signal=sigmaI/no_points;
	costs.newc.noise=sqrt(((sigmaI2-2*Ibar*sigmaI)/no_points)+Ibar*Ibar);
	}

void	Object3d::calc_change_bin(CHANGE_T ch){
int             i;
double         A,mag,temp;
double         r,p;
double         I,sigmaI=0,sigmaI2=0,Ibar=0;
complex <CGH_TYPE> 	amp;

	//Binary phase version a factor of 2 in the mag
	A=ch.mag*2;
	for (i = 0; i < no_points; i++) {
        	mag =A  / (r = rdist(position[i].x, position[i].y, ch.pos.x, ch.pos.y, depth[i]));
		//Then only require 1/2 as many sines and cosines
		p = phase(r) + ch.newphase;
		amp=newdata[i]=olddata[i]+polar(mag,p);
		I=real(amp*conj(amp));
		sigmaI+=I;             //sum the intensity
		sigmaI2+=(I*I);        //sum of intensity ^2;
        	}
	Ibar=costs.newc.signal=sigmaI/no_points;
	costs.newc.noise=sqrt(((sigmaI2-2*Ibar*sigmaI)/no_points)+Ibar*Ibar);
	}

void	Object3d::calc_change(CHANGE_T ch){
int             i;
double         mag,temp;
double         r,p1,p2;
double         I,sigmaI=0,sigmaI2=0,Ibar=0;
complex <CGH_TYPE> 	amp;

	//substitue in to allow vectorization
	for (i = 0; i < no_points; i++) {
		mag =ch.mag  / (r = rdist(position[i].x, position[i].y, ch.pos.x, ch.pos.y, depth[i]));
		temp = phase(r);
		p1 = temp+ch.newphase;
		p2 = temp+ch.oldphase;
		amp=newdata[i]=olddata[i]+polar(mag,p1)-polar(mag,p2);
		// The previous line MAY be inefficient...... matt 11/Aug/98
		I=real(amp*conj(amp));
		sigmaI+=I;             //sum the intensity
		sigmaI2+=(I*I);        //sum of intensity ^2;
		}
	Ibar=costs.newc.signal=sigmaI/no_points;
	costs.newc.noise=sqrt(((sigmaI2-2*Ibar*sigmaI)/no_points)+Ibar*Ibar);
	}
 
Object3d::Object3d(ifstream * is){
char temp[MAX_STRING],phase_table_name[MAX_STRING];
int counter=0;

	(*is)>>posname>>ampname;
	cout<<"pos name="<<posname<<" amp name="<<ampname<<"\n";
	do {
		(*is)>>temp;
		counter++;
		if (strcmp(temp, "imagesize") == 0) {
			(*is)>>no_points;
			}
		if (strcmp(temp, "distance") == 0) {
			cout<<"Warning `distance' statement found in 3dObject definition\n";
			}
		}while (strcmp(temp, "end") != 0&&counter<MAX_OPTIONS);
	if(counter==MAX_OPTIONS){
        	cout<<"3dOjbect option read error \n";
        	exit(1);
        	}
	
	position=	new POINT[no_points];
	newdata=	new complex <CGH_TYPE>[no_points];
	olddata=	new complex <CGH_TYPE>[no_points];
	depth=		new float[no_points];

	if(position==NULL||newdata==NULL||olddata==NULL)p("line mem alloc err");

	for(int i=0;i<no_points;i++){
        	newdata[i]=complex <CGH_TYPE>(0,0);
        	olddata[i]=complex <CGH_TYPE>(0,0);
		}
	}

void	Object3d::loaddata(void){
int i;

	ifstream is;
	is.open(posname);
	if(is.bad()){	cout<<"3dObject .pos data no good exiting \n";
			exit(3);
			}
	for(i=0;i<no_points;i++){
		is>>(position[i].x)>>(position[i].y)>>depth[i];
		}
	is.close();
	}

// now Object3dgs based on object3d, main difference is that the .pos
// data contains gray levels as well as the depth. For an explanation
// of the long-handidness of this see Object3d

typedef class Object3dgs : public Object3d {
	private:
		float		*gray_level;
		float		*depth;
		float 		total_intensity;
	public:
		void	           loaddata(void);
		void	           calc_change(CHANGE_T);
		void	           calc_change_bin(CHANGE_T);
		void	           intro_change(CHANGE_T); 
		Object3dgs(ifstream *is);
		Object3dgs(){};//default constructor
		// The cost parameters are kept raw here for use by the Image object
		// it consists of a COST_T structure
	} Object3dgs;



void	Object3dgs::intro_change(CHANGE_T ch){
int             i;
double         mag;
double         r,p;
double         I,sigmaI=0,sigmaI2=0,sigmaIw=0,Ibar=0,Ibarw=0;
complex <CGH_TYPE> 	amp;

for (i = 0; i < no_points; i++) {
		mag =ch.mag  / (r = rdist(position[i].x, position[i].y, ch.pos.x, ch.pos.y, depth[i]));
		//this means a positive phase is equivalent to extra path length
		p = phase(r) + ch.newphase;
		amp=newdata[i]=olddata[i]+polar(mag,p);
		I=real(amp*conj(amp));
//experimental weighting system.....
		sigmaI+=I;             //sum the intensity
		sigmaIw+=I/gray_level[i];
		sigmaI2+=I*I/gray_level[i];//sum of intensity ^2;
		}
	Ibarw=sigmaIw/no_points;
	Ibar=costs.newc.signal=sigmaI/no_points;
	//costs.newc.noise=sqrt(((sigmaI2-2*Ibarw*sigmaI)/no_points)+Ibarw*Ibarw);
	costs.newc.noise=sqrt(((sigmaI2-2*Ibar*sigmaI)/no_points)+Ibar*Ibar);
	}

void	Object3dgs::calc_change_bin(CHANGE_T ch){
int             i;
double         A,mag;
double         r,p;
double         I,sigmaI=0,sigmaI2=0,sigmaIw=0,Ibar=0,Ibarw=0;
complex <CGH_TYPE> 	amp;

	//Binary phase version a factor of 2 in the mag
	A=ch.mag*2;
	for (i = 0; i < no_points; i++) {
        	mag =A  / (r = rdist(position[i].x, position[i].y, ch.pos.x, ch.pos.y, depth[i]));
	//Then only require 1/2 as many sines and cosines
		p = phase(r) + ch.newphase;
		amp=newdata[i]=olddata[i]+polar(mag,p);
		I=real(amp*conj(amp));
		sigmaI+=I;             //sum the intensity
		sigmaIw+=I/gray_level[i];
		sigmaI2+=(I*I)/gray_level[i];        //sum of intensity ^2;
        	}
	Ibarw=sigmaIw/no_points;
	Ibar=costs.newc.signal=sigmaI/no_points;
	//costs.newc.noise=sqrt(((sigmaI2-2*Ibarw*sigmaI)/no_points)+Ibarw*Ibarw);
	costs.newc.noise=sqrt(((sigmaI2-2*Ibar*sigmaI)/no_points)+Ibar*Ibar);
	}

void	Object3dgs::calc_change(CHANGE_T ch){
int             i;
double         mag,temp;
double         r,p1,p2;
double         I,sigmaI=0,sigmaI2=0,sigmaIw=0,Ibar=0,Ibarw=0;
complex <CGH_TYPE> 	amp;	

	for (i = 0; i < no_points; i++) {
		mag =ch.mag  / (r = rdist(position[i].x, position[i].y, ch.pos.x, ch.pos.y, depth[i]));
		temp = phase(r);
		p1 = temp+ch.newphase;
		p2 = temp+ch.oldphase;
		amp=newdata[i]=olddata[i]+polar(mag,p1)-polar(mag,p2);
		I=real(amp*conj(amp));
		sigmaI+=I;             //sum the intensity
		sigmaIw+=I/gray_level[i];
		sigmaI2+=(I*I)/gray_level[i];        //sum of intensity ^2;
		}
	Ibarw=sigmaIw/no_points;
	Ibar=costs.newc.signal=sigmaI/no_points;
	//costs.newc.noise=sqrt(((sigmaI2-2*Ibarw*sigmaI)/no_points)+Ibarw*Ibarw);
	costs.newc.noise=sqrt(((sigmaI2-2*Ibar*sigmaI)/no_points)+Ibar*Ibar);
	}
 
Object3dgs::Object3dgs(ifstream * is){
char temp[MAX_STRING],phase_table_name[MAX_STRING];
int counter=0;

	(*is)>>posname>>ampname;
	cout<<"pos name="<<posname<<" amp name="<<ampname<<"\n";
	do {
		(*is)>>temp;
		counter++;
		if (strcmp(temp, "imagesize") == 0) {
			(*is)>>no_points;
			}
		if (strcmp(temp, "distance") == 0) {
			cout<<"Warning `distance' statement found in 3dObjectgs definition\n";
			}
		}while (strcmp(temp, "end") != 0&&counter<MAX_OPTIONS);
	if(counter==MAX_OPTIONS){
        	cout<<"gs3dOjbect option read error \n";
        	exit(1);
        	}
	
	position=	new POINT[no_points];
	newdata=	new complex <CGH_TYPE>[no_points];
	olddata=	new complex <CGH_TYPE>[no_points];
	depth=		new float[no_points];
	gray_level=	new float[no_points];
	if(position==NULL||newdata==NULL||olddata==NULL)p("line mem alloc err");
	for(int i=0;i<no_points;i++){
        	newdata[i]=complex <CGH_TYPE>(0,0);
        	olddata[i]=complex <CGH_TYPE>(0,0);
		}
	}

void	Object3dgs::loaddata(void){
	int i;
	float 	total=0;
	ifstream is;
	is.open(posname);
	if(is.bad()){	cout<<"3dObject .pos data no good exiting \n";
			exit(3);
			}
	for(i=0;i<no_points;i++){
		is>>(position[i].x)>>(position[i].y)>>depth[i]>>gray_level[i];
		}
	for(i=0,total=0;i<no_points;i++){
		total+=gray_level[i];
		}
	//normalize the gray_level so the average level is 1
	total=no_points/total;
	for(i=0;i<no_points;i++){
		gray_level[i]*=total;
		}
	is.close();
	}


//-------IMAGE---------------------image object declarations-- -- ----------
class Image:public Optics_Object{
	public:
        	COST_T 		costs;
		int             no_objects;
		Image_Object 	objects;
		Image(ifstream *);
		void	           loaddata(void);
		void	           savedata(void);
		COST            do_change(CHANGE_T);
		void	           do_report(ofstream *);
		void	           keep_change(void);
	};


Image::Image(ifstream * is){
char            temp[300];
int             counter=0,no_lines=0,no_patches=0,no_line_3ds=0,no_line_gs=0;
int		pimage=0, ppatch=0;
Image_Object 	*image;
int 		no_lines_read=0;

	no_objects=0;

	//set objects to point at the image base
	image=&objects;
	//read in the particulars about the image parts
	do {
		(*is)>>temp;
		no_lines_read++;
                if ((strcmp(temp, "3dObject") == 0)) {
			image->next=(void	*)new Object3d(is);
			if(image->next==NULL){
                                cout<<"Fatal error:Unable to allocate imagepart "<<pimage<<"\n";
                                exit(3);
                                }
			image=(Image_Object *)image->next;
			no_objects++;
			}
                if ((strcmp(temp, "3dgsObject") == 0)) {
			image->next=(void	*)new Object3dgs(is);
			if(image->next==NULL){
                                cout<<"Fatal error:Unable to allocate imagepart "<<pimage<<"\n";
                                exit(3);
                                }
			image=(Image_Object *)image->next;
			no_objects++;
			}
	                        
	                        
                if ((strcmp(temp, "2dObject") == 0)) {
			image->next=(void	*)new Object2d(is);
			if(image->next==NULL){
                                cout<<"Fatal error:Unable to allocate imagepart "<<pimage<<"\n";
                                exit(3);
                                }
			image=(Image_Object *)image->next;
			no_objects++;
			}
	                        
	}while ((strcmp(temp, "endcon")!=0)&&(no_lines_read<100000));
if(no_lines_read>99990){
	cout<<"Image::Image read options error, no 'endcon' statement found\n";
	}
if(no_objects==0){
	cout<<"Image::Image No image objects were found\n";
	exit(3);
	}
	cout<<"Image::Image "<<no_objects<<" Image objects were found\n";

	}

void	Image::loaddata(void) {
Image_Object	*image;

	//set the image to point at the first object
	image=(Image_Object *)objects.next;
	while(image!=NULL){
		image->loaddata();
		image=(Image_Object *)image->next;
		}
	}

void	Image::savedata(void) {
Image_Object    *image;

        //set the image to point at the first object
        image=(Image_Object *)objects.next;
        while(image!=NULL){
                image->savedata();
                image=(Image_Object *)image->next;
                }
	}


COST Image::do_change(CHANGE_T change){
COST cost;
Image_Object    *image;

        image=(Image_Object *)objects.next;
	costs.oldc.signal=0;
	costs.oldc.noise=0;
	costs.newc.signal=0;
	costs.newc.noise=0;
	while(image!=NULL){	
		costs=costs+image->do_change(change);
		image=(Image_Object*)image->next;
		}
	//cout<<"Image::do_change;"<<costs.newc.signal<<" "<<costs.newc.noise<<endl;
	cost.newcost=-a*costs.newc.signal+b*costs.newc.noise;
	cost.oldcost=-a*costs.oldc.signal+b*costs.oldc.noise;
	return cost;
	}

void	Image::do_report(ofstream * os){
Image_Object    *image;

        image=(Image_Object *)objects.next;
	while(image!=NULL){
		(*os)<<" ";image->do_report(os);
                image=(Image_Object *)image->next;
		}
	}


void	Image::keep_change(void){
Image_Object    *image;

        image=(Image_Object *)objects.next;
        while(image!=NULL){
		image->keep_change();
                image=(Image_Object *)image->next;
                }
	}
//----------------------------get_options----------------------------------
void	get_options(ifstream *,char *,int);

void	get_options(ifstream *is,char * rep_name,int *rep_gran){
char temp[MAX_NAME];
int counter=0;
		do{
		(*is)>>temp;
		counter++;
		if(strcmp(temp, "max_sweeps") == 0) {
                        (*is)>>max_sweeps;
                        }
		if(strcmp(temp,"report")==0) {
			(*is)>>rep_name;
			(*is)>>(*rep_gran);
			}
		if(strcmp(temp,"costa")==0){
			(*is)>>a;
			}
		if(strcmp(temp,"costb")==0){
			(*is)>>b;
			}
	}while (strcmp(temp, "endoptions") != 0&&counter<MAX_OPTIONS);
	if(counter==MAX_OPTIONS){
        	cout<<"get_options read error \n";
        	exit(1);
        	}
	}

//----------------------------------main-----------------------------------


main(int argc,char **argv){
int i=0,rep_count=0,rep_gran=2500;
long t_stamp=0;
char fname[MAX_NAME],rep_name[MAX_NAME];
CHANGE_T change;
COST icost;
ifstream            is;
ofstream            rep_o;

	void	catch_sigusr1(int signo);
	void	catch_sigusr2(int signo);	
	if (argc>1){
		strcpy(fname,argv[1]);
		}
	else{
		cout<<"DS_EXPT: ERROR: usage ds_expt controlfile\n";
		exit(3);
		}
	cout<<"ds_expt version 1.0 (Direct search program pruned version 3.0 1998 onwards)\n";
	cout<<"ds_expt version 1.0 (c) Matt Clark Sept 2002 released under GPL\n";
	is.open(fname);
	if(is.bad()){
		cout<<"error unable to open control file "<<fname<<"\n";
		exit(1);}
	cout<<"start get options \n";
	get_options(&is,rep_name,&rep_gran);
	cout<<" The initial cost factors are a="<<a<<" b="<<b<<"\n";
	cout<<" start setup hobject\n";
	Hologram        hologram(&is);
	Image           image(&is);
	if(rep_gran!=0){
		rep_o.open(rep_name);
		cout<<"reporting on\n";
		}
	is.close();
	image.loaddata();
	hologram.loaddata();
	icost.oldcost=1e300;//ensure the acceptance of the first change
	t_stamp=0;
	// set signal sigusr1 handeler so that the program dumps the data and then stops
	// set signal sigusr2 handeler so that the program dumps the data and continues
	signal(SIGUSR1,catch_sigusr1);
	signal(SIGUSR2,catch_sigusr2);
	cout<<"Entering main"<<endl;
	// main program loop starts here	
	do{     
		t_stamp++;
		rep_count++;
		change=hologram.generate_change();
		icost=image.do_change(change);
		// The first statement is the usual, the second makes sure that the program
		// turns ALL the pixels on
		if(	(icost.newcost<=icost.oldcost)||
			((change.oldphaselevel==NULL_PHASE)&&(hologram.sweep>1)))
			{
			hologram.keep_change();
			image.keep_change();
			icost.oldcost=icost.newcost;                    
			}       
		//Do the report file stuff
		if((rep_gran!=0)&&(rep_count>rep_gran)){
				rep_o<<" "<<t_stamp<<" ";
				rep_count=1;
				hologram.do_report(&rep_o);
				image.do_report(&rep_o);
				rep_o<<endl;
				}
		//Dump the data if sigusr1 or sigusr2  has been caught
		if(dump_data==TRUE){
		        hologram.savedata();
       			image.savedata();
			dump_data=FALSE;
			}
		//prevent the program from executing forever...
		if(hologram.sweep>max_sweeps){
			continue_ok=FALSE;
			}
		}
		while(!hologram.stable_solution()&&continue_ok==TRUE);
	hologram.savedata();
	image.savedata();
	if(rep_gran!=0){
		rep_o.close();
		}
	cout<<"end\n";
	return (0);
	}

void	catch_sigusr1(int ){
	cout<<"***** caught signal user 1, aborting and saving data *****\n";
	cerr<<"***** caught signal user 1, aborting and saving data *****\n";
	continue_ok=FALSE;
	}

void	catch_sigusr2(int){
	if(dump_data==FALSE){
		cout<<"***** caught signal user 2, saving data and continuing *****\n";
		cerr<<"***** caught signal user 2, saving data and continuing *****\n";
		}
		else{
		cout<<"***** caught signal user 2, waiting to start save cycle *****\n";
		} 
	dump_data=TRUE;
	}
